<?php
class parseGif256x192 {
	private $curScreen = array();	// Массив с данными текущего состояния экрана
	private $filler = array();		// Массив данных для "точечной" отправки на экран
	private $sequences = array();	// Последовательности для объединения в процедуры
	
	// Настройки
	var $procPrefix = 'x';
	var $outputDir = false;
	var $cycleAnimation = true;
	var $separateFrames = false;		// Каждый фрейм - заново (без учета предыдущего) 
	var $processSequences = array();	// Вычислять следующие последовательности

	// Результат
	var $totalstat = 0;
	var $framestats = array();		// Пофреймовая статистика обработки
	var $funcstats = 0;				// Статистика функций
	
	function __construct($config) {
		$this->procPrefix = isset($config['procPrefix']) ? $config['procPrefix'] : $this->procPrefix;
		$this->cycleAnimation = isset($config['cycleAnimation']) ? $config['cycleAnimation'] : $this->cycleAnimation;
		$this->outputDir = isset($config['outputDir']) ? $config['outputDir'] : $this->outputDir;
		$this->processSequences = isset($config['processSequences']) ? $config['processSequences'] : $this->processSequences;
		$this->separateFrames = isset($config['separateFrames']) ? $config['separateFrames'] : $this->separateFrames;
		
		// initial curScreen array
		for ($i =0; $i < 6144; $i++) $this->curScreen[$i] = 0;
	}	
	
	private function proceedFrame($frame) {
		$result = array();
		for ($y = 0; $y < 192; $y++) {
			for ($x = 0; $x < 32; $x++) {
				list($byte, $address) = $this->proceedByte($frame, $x, $y);
				if ($byte === false) continue;
				
				if (!isset($result[$byte])) $result[$byte] = array();
				$result[$byte][] = $address;
			}
		}
		
		return $result;
	}

	private function proceedByte($frame, $x, $y) {
		$byte = 0;
		for ($i = 0; $i < 8; $i ++) {
			if (!imagecolorat($frame, $x*8 + $i, $y)) continue;
				
			$byte += pow(2, 7-$i);
		}
	
		$d = ($y & 0xc0)*0x20 + ($y%8)*256 + ($y & 0x38) * 4 + $x;
	
		if ($this->curScreen[$d] == $byte) return array(false, false);
		
		$this->curScreen[$d] = $byte;
	
		return array($byte, $d);
	}	

	private function searchSequences($len = 8) {
		foreach ($this->filler as $key=>$data) {
			foreach($data as $byte=>$addresses) {
				$offset = 0;
				while ($offset++ < count($addresses) - $len) {
					$curSequence = array_slice($addresses, $offset, $len);
					if (in_array(-1, $curSequence)) continue; // Already has sequence
					
					if (!$result = $this->testSequence($curSequence)) continue;
						
					// Make sequence
					$sequenceName = $this->procPrefix.'s_'.substr(md5(serialize($curSequence)), 16);
					if (!isset($this->sequences[$sequenceName])) $this->sequences[$sequenceName] = array();
					
					foreach ($result as $r) {
						$this->sequences[$sequenceName][] = $r; 
							
						for ($i = 0; $i < $len; $i++) {
							$this->filler[$r['key']][$r['byte']][$r['offset'] + $i] = -1;
						}
					}
					break;
				}
			}
		}
	}

	private function testSequence($sequence = array()) {
		$len = count($sequence);
		$result = array();
		foreach ($this->filler as $key=>$data) {
			foreach($data as $byte=>$addresses) {
				$offset = 0;
				while ($offset++ < count($addresses) - $len) {
					
					$is_sequence = true;
					for ($i = 0; $i < $len; $i++) {
						if ($sequence[$i] != $addresses[$offset + $i]) {
							$is_sequence = false;
							break;
						}
					}
						
					if ($is_sequence) {
						$result[] = array('key' => $key, 'byte' => $byte, 'offset' => $offset, 'sequence' => $sequence);
						break;
					}
				}
			}
		}
		
		return count($result) > 1 ? $result : false;
	}
	
	private function generateCode($durations) {
		$generated = array('frames' => array(), 'functions' => false);
		
		foreach ($this->processSequences as $i) {
			$this->searchSequences($i);
		}
		
		// Remove processed in sequences bytes from filler
		foreach ($this->filler as $key=>$data) {
			foreach($data as $byte=>$addresses) {
				foreach($addresses as $key2=>$address) {
					if ($address == -1) {
						unset($this->filler[$key][$byte][$key2]);
					}
				}
			}
		}
		
		// Generate functions
		$output = '';
		foreach ($this->sequences as $fnmame=>$s) {
			$output .= $fnmame."\n";
			
			$s = reset($s);
			$output .= $this->generateSequenceCode($s['sequence']);
			$output .= "\t\tret\n";
			$this->funcstats ++;
		}
		// Save functions
		if ($this->outputDir) {
			$fp = fopen($this->outputDir.'/'.$this->procPrefix.'func.asm', 'w');
			fwrite($fp, $output);
			fclose($fp);
		}
		else {
			$generated['functions'] = array('filename' => $this->procPrefix.'func.asm', 'content' => $output);
		}
		
		foreach ($this->filler as $key=>$data) {
			$output = $this->procPrefix.$key;
			$this->framestats[$key]['bytes'] = 0;
			
			$curA = -1;
			foreach($data as $byte=>$addresses) {
				if ($byte == 0) {
					$output .= "\t\txor a\n";
					$this->framestats[$key]['bytes']++;
				}
				elseif ($byte - $curA == 1) {
					$output .= "\t\tinc a\n";
					$this->framestats[$key]['bytes']++;
				}
				else {
					$output .= "\t\tld a, ".$byte."\n";
					$this->framestats[$key]['bytes'] += 2;
				}
				$curA = $byte;
				
				// Call sequences
				foreach ($this->sequences as $fnmame=>$i) foreach ($i as $s) {
					if ($s['byte'] == $byte && $s['key'] == $key) {
						$output .= "\t\tcall ".$fnmame."\n";
						$this->framestats[$key]['bytes'] += 3;
					}
				}

				$output .= $this->generateSeriaV($addresses, $this->framestats[$key]['bytes']);
				$output .= $this->generateSeriaH($addresses, $this->framestats[$key]['bytes']);
				
				// Формируем оставшиеся фрагменты:
				// ld (addr1), a
				foreach ($addresses as $address) {
					$output .= "\t\tld (#".dechex(0x4000 + $address)."), a\n";
					$this->framestats[$key]['bytes'] += 3;
				}
			}
			
			$output .= "\t\tret\n";
			$this->framestats[$key]['bytes']++;
			
			// Save result
			if ($this->outputDir) {
				$fp = fopen($this->outputDir.'/'.$this->procPrefix.$key.'.asm', 'w');
				fwrite($fp, $output);
				fclose($fp);
			}
			else {
				$duration = intval($durations[$key] / 2);
				
				$generated['frames'][] = array(
					'duration' => $duration < 1 ? 1 : ($duration > 255 ? 255 : $duration),
					'filename' => $this->procPrefix.$key.'.asm', 
					'function_name' => $this->procPrefix.$key,
					'content' => $output
				);
			}
		}
		
		return $this->outputDir ? true : $generated; 
	}
	
	private function generateSequenceCode($sequence) {
		$output = $this->generateSeriaV($sequence, $this->funcstats);
		$output .= $this->generateSeriaH($sequence, $this->funcstats);
		
		// Формируем оставшиеся фрагменты:
		// ld (addr1), a
		foreach ($sequence as $address) {
			$output .= "\t\tld (#".dechex(0x4000 + $address)."), a\n";
			$this->funcstats += 3;
		}
		
		return $output; 
	}
	
	/* Формирует из массива $source фрагменты типа:
	 * ld hl, addr : ld (hl), a : inc h : ld (hl), a 
	 */
	private function generateSeriaV(&$source, &$counter) {
		sort($source);
		
		$output = '';
		
		foreach ($source as $address) {
			$seria = array();
			
			while(true) {
				if (!in_array($address, $source)) break;
				
				$seria[] = $address;
				$address += 0x100;
			}

			// Seria end
			if (count($seria) > 2) {
				// Remove seria addresses from source
				foreach ($source as $k=>$a) {
					if (in_array($a, $seria)) unset($source[$k]);
				}

				$output .= "\t\tld hl, #".dechex(0x4000 + array_shift($seria))."\n";
				$output .= "\t\tld (hl), a\n";
				$counter += 4;
					
				foreach($seria as $foo) {
					$output .= "\t\tinc h\n";
					$output .= "\t\tld (hl), a\n";
					$counter += 2;
				}
			}
		}
		
		return $output;
	}
	
	/* Формирует из массива $source фрагменты типа:
	 * ld hl, addr : ld (hl), a : inc hl : ld (hl), a
	*/
	private function generateSeriaH(&$source, &$counter) {
		sort($source);
		
		$output = '';
		$processedDE = false;
		
		foreach ($source as $address) {
			$seria = array();
				
			while(true) {
				if (!in_array($address, $source)) break;
	
				$seria[] = $address;
				$address++;
			}
	
			// Seria end
			if (count($seria) >= 2) {
				// Remove seria addresses from source
				foreach ($source as $k=>$a) {
					if (in_array($a, $seria)) unset($source[$k]);
				}
	
				if (count($seria) == 2) {
					if (!$processedDE) {
						$output .= "\t\tld d,a\n";
						$output .= "\t\tld e,a\n";
						$counter += 2;
						$processedDE = true;
					}
				
					$output .= "\t\tld (#".dechex(0x4000 + $seria[0])."), de\n";
					$counter += 4;
				}
				else {
					$output .= "\t\tld hl, #".dechex(0x4000 + array_shift($seria))."\n";
					$output .= "\t\tld (hl), a\n";
					$counter += 4;
				
				
					foreach($seria as $foo) {
						$output .= "\t\tinc hl\n";
						$output .= "\t\tld (hl), a\n";
						$counter += 2;
					}
				}
			}
		}
	
		return $output;
	}
		
	function parseGIF($filename) {
		require_once(dirname(__FILE__).'/GifFrameExtractor.php');
		if (!GifFrameExtractor::isAnimatedGif($filename)) return false;
		
		$gfe = new GifFrameExtractor();
		$gfe->extract($filename);
		$frames = $gfe->getFrameImages();
		$durations = $gfe->getFrameDurations();
		
		foreach ($frames as $key=>$frame) {
			$this->filler[$key] = $this->proceedFrame($frame);
			
			if ($this->separateFrames) {
				for ($i =0; $i < 6144; $i++) $this->curScreen[$i] = 0;
			}
		}
		
		if ($this->cycleAnimation) {
			$this->filler[$key + 1] = $this->proceedFrame($frames[0]);
			$durations[] = $durations[0];
		}
		
		$result = $this->generateCode($durations);
		
		// Calculate total bytes
		$this->totalstat = $this->funcstats;
		foreach ($this->framestats as $frame=>$stat) $this->totalstat += $stat['bytes'];
		
		return $result;
	}
}